/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
package com.openexchange.office.rt2.ws;

import java.util.ArrayList;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.BooleanUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.openexchange.ajax.SessionUtility;
import com.openexchange.config.ConfigurationService;
import com.openexchange.configuration.CookieHashSource;
import com.openexchange.configuration.ServerConfig.Property;
import com.openexchange.exception.OXException;
import com.openexchange.office.rt2.exception.RT2SessionInvalidException;
import com.openexchange.office.rt2.exception.RT2TypedException;
import com.openexchange.office.rt2.protocol.RT2Message;
import com.openexchange.office.rt2.protocol.RT2MessageGetSet;
import com.openexchange.office.rt2.protocol.RT2Protocol;
import com.openexchange.office.rt2.protocol.value.RT2MessageType;
import com.openexchange.office.session.SessionService;
import com.openexchange.office.tools.osgi.ServiceLookupRegistryService;
import com.openexchange.session.Session;

public class RT2SessionValidator {	
	
	private static final Logger log = LoggerFactory.getLogger(RT2SessionValidator.class);
	
	private final ConfigurationService configService;
	private final SessionService sessionService;
	
	public RT2SessionValidator() {
		this.configService = ServiceLookupRegistryService.getInstance().getService(ConfigurationService.class);
		this.sessionService = ServiceLookupRegistryService.getInstance().getService(SessionService.class);
	}

	/**
	 * Will refresh current session of webSocket, but will fail if session is already expired or invalid.
	 * @param rt2Msg Current request
	 * @param webSocket The connection to check the session of
	 */
	public void validateSession(RT2Message rt2Msg, RT2EnhDefaultWebSocket webSocket) throws Exception {
	    try {
	        impl_refreshOrFailOnInvalidSession (rt2Msg, webSocket);
	    } catch (RT2SessionInvalidException ex) {
	        // Handle invalid exception and check out if the request can
	        // be ignored or we have to process the exception. In that
	        // case the handler has to re-throw the original exception.
	        impl_handleInvalidSession(rt2Msg, webSocket, ex);
	    }		
	}
	
    private void impl_refreshOrFailOnInvalidSession (final RT2Message aRequest, final RT2EnhDefaultWebSocket webSocket)
            throws Exception
        {
            log.debug("impl_refreshOrFailOnInvalidSession");

            if ( ! aRequest.isSessionMessage())
                return;

            // TODO: make this sense? security issue if this can be disabled via header!
            final Boolean bDBGDisableSessionValidation = RT2MessageGetSet.getOptionalHeader(aRequest, RT2Protocol.HEADER_DEBUG_DISABLE_SESSION_VALIDATION, Boolean.FALSE);

            if (BooleanUtils.isTrue(bDBGDisableSessionValidation)) {
                log.debug("RT2: session validation disabled by request [{}]", aRequest);
                return;
            }

            if (aRequest.getSessionID() == null) {
                final String logMessage = "no session defined for request [" + aRequest + "]";
                final RT2TypedException ex = new RT2SessionInvalidException(new ArrayList<>(), "details", logMessage);
                throw ex;
            }

            final Session aSession = sessionService.getSession4Id(aRequest.getSessionID().getValue());
            if (aSession == null) {
                final String logMessage = "RT2: Session '" + aRequest.getSessionID() + "' do not exists - or is expired. Request: [" + aRequest + "]";
                final RT2TypedException ex = new RT2SessionInvalidException(new ArrayList<>(), "details", logMessage);
                throw ex;
            }

            log.trace("Session found for request with session id " + aRequest.getSessionID());

            final boolean bIsValid = sessionService.isSessionValid(aSession);
            if ( ! bIsValid) {
                final String logMessage = "RT2: Session '"+aSession.getSessionID()+"' is not valid. Request: [" + aRequest + "]";
                final RT2TypedException ex = new RT2SessionInvalidException(new ArrayList<>(), "details", logMessage);
                throw ex;
            }

            impl_validateSessionSecrets(aRequest, aSession, webSocket);
        }

        //-------------------------------------------------------------------------
        private void impl_validateSessionSecrets (final RT2Message aRequest, final Session aSession, final RT2EnhDefaultWebSocket webSocket) throws RT2TypedException {
        	
        	Object obj = webSocket.getUpgradeRequestSessionValid();
        	if (obj == null) {
        		obj = Boolean.TRUE;
        	}
        	Boolean sessionValid = (Boolean) obj;
            if (!sessionValid.booleanValue()) {
                final String logMessage = "RT2: Upgrade request has no valid session '" + aSession.getSessionID() + "'";
                final RT2TypedException ex = new RT2SessionInvalidException(new ArrayList<>(), "details", logMessage);
                throw ex;
            }

            final String               sCookieHash     = configService.getProperty(Property.COOKIE_HASH.getPropertyName());
            final CookieHashSource     aHashSource     = CookieHashSource.parse(sCookieHash);
            final HttpServletRequest   aServletRequest = webSocket.getUpgradeRequest();

            try {
            	final UpgradeRequestWrapper requestWrapped = new UpgradeRequestWrapper(aServletRequest, true);
                SessionUtility.checkSecret(aHashSource, requestWrapped, aSession);
                webSocket.setUpgradeRequestSessionValid(true);
            }
            catch (OXException ex) {
            	webSocket.setUpgradeRequestSessionValid(false);
                final String logMessage = "RT2: Session secret validation failed for session '" + aSession + "'";
                final RT2TypedException rt2Ex = new RT2SessionInvalidException(new ArrayList<>(), "details", logMessage);
                throw rt2Ex;
            }
            catch (Throwable ex) {
            	webSocket.setUpgradeRequestSessionValid(false);
                com.openexchange.exception.ExceptionUtils.handleThrowable(ex);
                log.error(ex.getMessage(), ex);
				throw ex;
        	}
        }

        //-------------------------------------------------------------------------
        private void impl_handleInvalidSession(RT2Message aRequest, final RT2EnhDefaultWebSocket webSocket, final RT2SessionInvalidException origEx) throws RT2SessionInvalidException {
            final RT2MessageType type = aRequest.getType();

            // In case we receive a ack or nack and the session is already
            // not valid anymore - we assume that this could be a time-out
            // handling of the client, where is closes the application and
            // logs, but the ACK/NACK msgs are still on there way.
            // So we have a special handling for these two messages, which
            // will be ignored.
            // ATTENTION: Never let messages pass if we know that the session
            // is not valid.
            if ((type == RT2MessageType.ACK_SIMPLE) ||
                (type == RT2MessageType.NACK_SIMPLE)) {
            	origEx.disableLog();
                throw origEx;
            }
            if (!webSocket.isUnavailabeLogged()) {
            	log.info("RT2: websocket channel {} received more requests from client {} while session is invalid - drop request", webSocket.getId(), aRequest.getClientUID());
            	webSocket.setUnavailabeLogged(true);
            }
            throw origEx;
        }
}
