/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.spellchecker.hunspell;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.regex.Pattern;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.openexchange.config.ConfigurationService;
import com.openexchange.office.spellchecker.api.ISpellChecker;
import com.openexchange.office.tools.common.string.UnicodeValidator;
import com.openexchange.spellchecker.hunspell.jna.Hunspell;

/**
 * {@link SpellChecker}
 *
 * @author <a href="mailto:oliver.specht@open-xchange.com">Oliver Specht</a>
 */
public class SpellChecker implements ISpellChecker {

    final private static Logger LOG = LoggerFactory.getLogger(SpellChecker.class);

    final private static String SEPARATOR_PATTERN = "[\t\n\u00a0\f\r \\.,;\\?!\":(){}\\[\\] ]";
    final private static String THE_DOT = ".";
    final private static char THE_BLANK = ' ';
    final private static int INIT_HUNSPELL_TIMER_PERIOD_MILLIS = 300000;

    /**
     * {@link HUNSPELL_STATUS}
     *
     * @author <a href="mailto:kai.ahrens@open-xchange.com">Kai Ahrens</a>
     * @since v7.10.2
     */
    private static enum HunspellStatus {
        ENABLED,
        DISABLED,
        INITIALIZE
    }

    /**
     * Initializes a new {@link SpellChecker}.
     */
    public SpellChecker( final ConfigurationService configService ){
        if (null != (m_configService = configService)) {
            boolean spellcheckingEnabled = configService.getBoolProperty("io.ox/office//module/spellingEnabled", false);

            LOG.info(new StringBuilder(128).
                append("Spellchecking is ").
                append(spellcheckingEnabled ? "enabled" : "disabled").
                append(" by configuration").toString());

            if (spellcheckingEnabled) {
                (m_initHunspellTimer = new Timer()).scheduleAtFixedRate(new TimerTask() {

                    @Override
                    public void run() {
                        if (HunspellStatus.ENABLED == (m_hunspellStatus = implInitHunspell())) {
                            implShutdownInitHunspellTimer();
                        }
                    }
                }, 0, INIT_HUNSPELL_TIMER_PERIOD_MILLIS);
            }
        }

    }

    /* (non-Javadoc)
     * @see com.openexchange.oxospellchecker.ISpellChecker#checkSpelling(java.lang.String, java.lang.String)
     */
    @Override
    public JSONArray checkSpelling(String text, String locale, int offset) {
        String textToUse = text + ' ';

        final JSONArray result = new JSONArray();
        final Hunspell.Dictionary dict = m_dictionaries != null ? m_dictionaries.getDictionary( locale ) : null;
        if( dict != null ){
            int startPos = 0;
            int len = 0;
            for( int pos = 0; pos < textToUse.length(); ++pos){
                final String substring = textToUse.substring(pos, pos + 1);
                if( Pattern.matches(SEPARATOR_PATTERN, substring) ){
                    if( len > 0 ){
                        // #40631# A dot might also be used for abbreviations.
                        final int extra = substring.equals(THE_DOT) ? 1 : 0;
                        final String word = textToUse.substring(startPos, startPos + len + extra);
                        // #65073# workaround for possible hunspell crash with composed unicode characters
                        // we just replace high/low-surrogate characters with a replacement character.
                        final String safeWord = UnicodeValidator.replaceComposedUnicodeCharacters(word, THE_BLANK);

                        if(dict.misspelled(safeWord)){
                            final JSONObject error = new JSONObject();
                            try {
                                error.put("start", startPos + offset);
                                error.put("length", len);
                                error.put("locale", dict.getLocale());
                                error.put("word", word);
                                final ListIterator<String> it = dict.suggest(safeWord).listIterator();
                                if(it.hasNext()){
                                    final JSONArray replacements = new JSONArray();
                                    while( it.hasNext() ){
                                        replacements.put( it.next() );
                                    }
                                    error.put("replacements", replacements);
                                }

                            } catch (@SuppressWarnings("unused") final JSONException e) {
                                // OK
                            }
                            result.put( error );
                        }
                        len = 0;
                    }
                    startPos = pos + 1;

                } else {
                    ++len;
                }
            }
        }
        return result;
    }
    /**
     * @param paragraph contains the parts of the paragraph combined to it's locale.
     * @param locale
     * @return array containing the suggestions. If the word is detected as correct the array contains the correct word, only
     */
    @Override
    public JSONArray checkParagraphSpelling( JSONArray paragraph ){
        final JSONArray  result = new JSONArray ();
        int pos = 0;
        int offset = 0;
        final StringBuilder text = new StringBuilder();
        String locale = null;
        try {
            while(pos < paragraph.length() ){
                final JSONObject element = paragraph.getJSONObject(pos);
                if( (locale != null && locale.compareTo( element.getString("locale")) != 0) ){
                    final JSONArray check = checkSpelling( text.toString(), locale, offset );
                    for( int i = 0; i < check.length(); ++i) {
                        result.put( check.get(i) );
                    }
                    offset += text.length();
                    text.setLength(0);
                    locale = null;
                }
                locale = element.getString("locale");
                text.append(element.getString("word"));
                ++pos;
            }
            if( text.length() > 0 ){
                final JSONArray check = checkSpelling( text.toString(), locale, offset );
                for( int i = 0; i < check.length(); ++i) {
                    result.put( check.get(i) );
                }
            }
        } catch (@SuppressWarnings("unused") final JSONException e) {
            // OK
        }
        return result;
    }

    /* (non-Javadoc)
     * @param word
     * @param locale
     * @return a list of suggestions.
     * @see com.openexchange.office.spellchecker.ISpellChecker#suggestReplacement(java.lang.String, java.lang.String)
     */
    @Override
    public List<String> suggestReplacements( String word, String locale ){
        final List<String>  result = new ArrayList<>();
        final Hunspell.Dictionary dict = m_dictionaries != null ? m_dictionaries.getDictionary( locale ) : null;
        if( dict != null ){
            final String safeWord = UnicodeValidator.replaceComposedUnicodeCharacters(word, THE_BLANK);
            final ListIterator<String> it = dict.suggest(safeWord).listIterator();
            while( it.hasNext() ){
                result.add( it.next() );
            }
        }
        return result;
    }
    /* (non-Javadoc)
     * @param word the word that has to be checked
     * @param locale the related locale
     * @return true if the word is detected as wrong. false if no dictionary for this locale is available
     * @see com.openexchange.office.spellchecker.ISpellChecker#isMisspelled(java.lang.String, java.lang.String)
     */
    @Override
    public boolean isMisspelled( String word, String locale ){
        final Hunspell.Dictionary dict = m_dictionaries != null ? m_dictionaries.getDictionary( locale ) : null;
        final String safeWord = UnicodeValidator.replaceComposedUnicodeCharacters(word, THE_BLANK);
        return dict != null ? dict.misspelled(safeWord) : false;
    }
    /* (non-Javadoc)
     * @return the list of supported locales
     * @see com.openexchange.office.spellchecker.ISpellChecker#getSupportedLocales()
     */
    @Override
    public List<String> getSupportedLocales(){
        return m_dictionaries != null ? m_dictionaries.getLocales() : new ArrayList<>();
    }

    public void shutdown() {
        implShutdownInitHunspellTimer();
    }

    // - Implementation --------------------------------------------------------

    /**
     *
     */
    protected synchronized void implShutdownInitHunspellTimer() {
        if (null != m_initHunspellTimer) {
            m_initHunspellTimer.cancel();
            m_initHunspellTimer = null;
        }
    }

    /**
     * @return Either {@link HunspellStatus.ENABLED} or {@link HunspellStatus.INITIALIZE}
     *  if Hunspell still needs to be initialized
     */
    protected synchronized HunspellStatus implInitHunspell() {
        try {
            String hunspellLibrary = m_configService.getProperty( "com.openexchange.office.spellchecker.hunspell.library", null);
            if(hunspellLibrary==null) {
                hunspellLibrary = m_configService.getProperty( "com.openexchange.spellchecker.hunspell.library", null);
            }
            if(hunspellLibrary==null) {
                for(String _hunspellLibrary:hunspellLibraryList) {
                    if(new File(_hunspellLibrary).canRead()) {
                        hunspellLibrary = _hunspellLibrary;
                        break;
                    }
                }
            }
            String dictionaryPath = m_configService.getProperty( "com.openexchange.office.spellchecker.hunspell.dictionaries", null);
            if(dictionaryPath==null) {
                dictionaryPath = m_configService.getProperty( "com.openexchange.spellchecker.hunspell.dictionaries", null);
            }
            if(dictionaryPath==null) {
                for(String _dictionaryPath:dictionaryPathList) {
                    if(new File(_dictionaryPath).canRead()) {
                        dictionaryPath = _dictionaryPath;
                        break;
                    }
                }
            }
            if(hunspellLibrary!=null && dictionaryPath!=null) {
                try{
                    final boolean hunspellLibExists = (new File(hunspellLibrary)).canRead();
                    final boolean dictionaryPathExists = (new File(dictionaryPath)).canRead();

                    if (hunspellLibExists && dictionaryPathExists){
                        m_hunspell = Hunspell.getInstance(hunspellLibrary);
                        m_dictionaries = new DictionaryContainer(m_hunspell, dictionaryPath);
                    }
                }
                catch(final UnsatisfiedLinkError e) {
                    if (LOG.isErrorEnabled()) {
                        LOG.error("Error initializing Hunspell - Unsatisfied Link Error", e);
                    }
                }
                catch(final Throwable e) {
                    if (LOG.isErrorEnabled()) {
                        LOG.error("Error initializing Hunspell", e);
                    }
                }
            }

            if (null!=m_hunspell && null!= m_dictionaries) {
                LOG.info("Spellchecker sucessfully initialized Hunspell library");
                return HunspellStatus.ENABLED;
            }

            if ((HunspellStatus.DISABLED == m_hunspellStatus)) {
                if (null == m_hunspell) {
                    LOG.warn("Spellchecker failed to initialize Hunspell library. Hunspell library cannot be found or executed: " + hunspellLibrary);
                }

                if (null == m_dictionaries) {
                    LOG.warn("Spellchecker failed to initialize Hunspell library. Hunspell dictionary path cannot be found or read: " + dictionaryPath);
                }

                LOG.info(new StringBuilder(128).
                    append("Spellchecker automatically retries to initialize Hunspell library with a period of ").
                    append(INIT_HUNSPELL_TIMER_PERIOD_MILLIS / 1000).append("s...").toString());
            }
            return HunspellStatus.INITIALIZE;
        }
        catch(java.security.AccessControlException e) {
            LOG.warn("AccessControlException: Spellchecker failed to initialize Hunspell library, please configure com.openexchange.spellchecker.hunspell.library and com.openexchange.spellchecker.hunspell.dictionaries.");
            return HunspellStatus.DISABLED;
        }
    }


    // - Members ---------------------------------------------------------------

    final private ConfigurationService m_configService;

    private Hunspell m_hunspell = null;

    private DictionaryContainer m_dictionaries = null;

    protected Timer m_initHunspellTimer = null;

    protected HunspellStatus m_hunspellStatus = HunspellStatus.DISABLED;

    private String[] hunspellLibraryList = {
        "/usr/lib/x86_64-linux-gnu/libhunspell-1.7.so.0",
        "/usr/lib64/libhunspell-1.7.so.0",
        "/usr/lib/x86_64-linux-gnu/libhunspell-1.4.so.0",
        "/usr/lib64/libhunspell-1.4.so.0",
        "/usr/lib/x86_64-linux-gnu/libhunspell-1.3.so.0",
        "/usr/lib64/libhunspell-1.3.so.0",
        "/usr/lib/x86_64-linux-gnu/libhunspell-1.2.so.0",
        "/usr/lib64/libhunspell-1.2.so.0"
    };

    private String[] dictionaryPathList = {
        "/usr/share/hunspell",
        "/usr/share/myspell"
    };
}
